@@ -13,6 +13,8 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+from midonet.neutron._i18n import _, _LE, _LW
+from midonet.neutron.common import utils as c_utils
 from midonet.neutron.db import agent_membership_db as am_db
 from midonet.neutron.db import port_binding_db as pb_db
 from midonet.neutron.db import provider_network_db as pnet_db
@@ -27,18 +29,14 @@ from neutron.extensions import extra_dhcp_opt as edo_ext
 from neutron.extensions import portsecurity as psec
 from neutron.extensions import providernet as pnet
 from neutron.extensions import securitygroup as ext_sg
-from neutron import i18n
 from neutron import manager
 from neutron.plugins.common import constants as service_constants
-from oslo_db import api as oslo_db_api
 from oslo_db import exception as oslo_db_exc
 from oslo_log import log as logging
 from oslo_utils import excutils
 
 
 LOG = logging.getLogger(__name__)
-_LE = i18n._LE
-_LW = i18n._LW
 
 
 class MidonetPluginV2(plugin.MidonetMixinBase,
@@ -52,6 +50,8 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
     # probably after a cycle. (for M release)
     # This plugin is for MidoNet v5.0 and later, where we have an
     # alternative metadata proxy which doesn't require Neutron agents.
+    # NOTE(yamamoto): While the order in this list doesn't matter
+    # functionality-wise, it's alphabetically sorted for easier comparison.
     supported_extension_aliases = [
         'agent',
         'agent-membership',
@@ -63,7 +63,8 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
         'port-security',
         'provider',
         'quotas',
-        'security-group'
+        'security-group',
+        'subnet_allocation',
     ]
 
     __native_bulk_support = True
@@ -76,7 +77,7 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
         LOG.debug('MidonetPluginV2.create_network called: network=%r', network)
 
         net_data = network['network']
-        tenant_id = self._get_tenant_id_for_create(context, net_data)
+        tenant_id = net_data['tenant_id']
         net_data['tenant_id'] = tenant_id
         self._ensure_default_security_group(context, tenant_id)
 
@@ -104,6 +105,11 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
                     LOG.exception(_LE("Failed to delete network %s"),
                                   net['id'])
 
+        # NOTE(kevinbenton): this extra lookup is necessary to get the
+        # latest db model for the extension functions
+        net_model = self._get_network(context, net['id'])
+        self._apply_dict_extend_functions('networks', net, net_model)
+
         LOG.debug("MidonetPluginV2.create_network exiting: net=%r", net)
         return net
 
@@ -158,13 +164,11 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
         LOG.debug("MidonetPluginV2.update_network exiting: net=%r", net)
         return net
 
-    @oslo_db_api.wrap_db_retry(max_retries=3, retry_interval=1,
-                               retry_on_request=True,
-                               retry_on_deadlock=True)
     def delete_network(self, context, id):
         LOG.debug("MidonetPluginV2.delete_network called: id=%r", id)
 
         with context.session.begin(subtransactions=True):
+            c_utils.check_delete_network_precommit(context, id)
             self._process_l3_delete(context, id)
             try:
                 super(MidonetPluginV2, self).delete_network(context, id)
@@ -232,7 +236,12 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
         port_data = port['port']
         # REVISIT(yamamoto): this nested transaction is a workaround
         # for bug #1490917.
-        with db_api.autonested_transaction(context.session):
+        with db_api.exc_to_retry(oslo_db_exc.DBDuplicateEntry),\
+                db_api.autonested_transaction(context.session):
+            # Set status along admin_state_up if the parameter is specified.
+            if port['port'].get('admin_state_up') is not None:
+                if not port['port']['admin_state_up']:
+                    port['port']['status'] = n_const.PORT_STATUS_DOWN
             # Create a Neutron port
             new_port = super(MidonetPluginV2, self).create_port(context, port)
 
@@ -303,6 +312,11 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
                     LOG.exception(_LE("Failed to delete port %s"),
                                   new_port['id'])
 
+        # NOTE(kevinbenton): this extra lookup is necessary to get the
+        # latest db model for the extension functions
+        port_model = self._get_port(context, new_port['id'])
+        self._apply_dict_extend_functions('ports', new_port, port_model)
+
         LOG.debug("MidonetPluginV2.create_port exiting: port=%r", new_port)
         return new_port
 
@@ -334,11 +348,13 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
         LOG.debug("MidonetPluginV2.update_port called: id=%(id)s "
                   "port=%(port)r", {'id': id, 'port': port})
 
-        with context.session.begin(subtransactions=True):
+        with db_api.exc_to_retry(oslo_db_exc.DBDuplicateEntry),\
+                context.session.begin(subtransactions=True):
 
             # update the port DB
             original_port = super(MidonetPluginV2, self).get_port(context, id)
             p = super(MidonetPluginV2, self).update_port(context, id, port)
+            c_utils.check_update_port(original_port, p)
 
             has_sg = self._check_update_has_security_groups(port)
             delete_sg = self._check_update_deletes_security_groups(port)
@@ -368,7 +384,27 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
 
             self.client.update_port_precommit(context, id, p)
 
-        self.client.update_port_postcommit(id, p)
+        # Set status along admin_state_up to update.
+        if p['admin_state_up']:
+            update_status = n_const.PORT_STATUS_ACTIVE
+        else:
+            update_status = n_const.PORT_STATUS_DOWN
+
+        try:
+            self.client.update_port_postcommit(id, p)
+            data = {'port': {'status': update_status}}
+            super(MidonetPluginV2, self).update_port(context, id, data)
+            # prevent binding:profile information from lacking
+            p['status'] = data['port']['status']
+        except Exception as ex:
+            with excutils.save_and_reraise_exception():
+                LOG.error(_LE("Failed to update a port %(port_id)s in "
+                              "MidoNet: %(err)s"), {"port_id": id, "err": ex})
+                try:
+                    data = {'port': {'status': n_const.PORT_STATUS_ERROR}}
+                    super(MidonetPluginV2, self).update_port(context, id, data)
+                except Exception:
+                    LOG.exception(_LE("Failed to update port status %s"), id)
 
         LOG.debug("MidonetPluginV2.update_port exiting: p=%r", p)
         return p
@@ -380,7 +416,7 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
                   {'security_group': security_group, 'default_sg': default_sg})
 
         sg = security_group.get('security_group')
-        tenant_id = self._get_tenant_id_for_create(context, sg)
+        tenant_id = sg['tenant_id']
         if not default_sg:
             self._ensure_default_security_group(context, tenant_id)
 
